@@ -29,6 +29,7 @@
 #include "openvswitch/vlog.h"
 #include "ovsdb-error.h"
 #include "ovsdb.h"
+#include "ovs-thread.h"
 #include "row.h"
 #include "storage.h"
 #include "table.h"
@@ -40,8 +41,8 @@ struct ovsdb_txn {
     struct ovsdb *db;
     struct ovs_list txn_tables; /* Contains "struct ovsdb_txn_table"s. */
     struct ds comment;
-    struct uuid txnid; /* For clustered mode only. It is the eid. */
-    size_t n_atoms;    /* Number of atoms in all transaction rows. */
+    struct uuid txnid; /* For clustered and relay modes.  It is the eid. */
+    size_t n_atoms;    /* Number of atoms in all changed transaction rows. */
     ssize_t n_atoms_diff;  /* Difference between number of added and
                             * removed atoms. */
 };
@@ -102,6 +103,7 @@ struct ovsdb_txn_row {
 static struct ovsdb_error * OVS_WARN_UNUSED_RESULT
 delete_garbage_row(struct ovsdb_txn *txn, struct ovsdb_txn_row *r);
 static void ovsdb_txn_row_prefree(struct ovsdb_txn_row *);
+static void ovsdb_txn_row_log(const struct ovsdb_txn_row *);
 static struct ovsdb_error * OVS_WARN_UNUSED_RESULT
 for_each_txn_row(struct ovsdb_txn *txn,
                       struct ovsdb_error *(*)(struct ovsdb_txn *,
@@ -111,6 +113,11 @@ for_each_txn_row(struct ovsdb_txn *txn,
  * processed.  */
 static unsigned int serial;
 
+/* Used by ovsdb_txn_row_log() to avoid reallocating dynamic strings
+ * every time a row operation is logged.
+ */
+DEFINE_STATIC_PER_THREAD_DATA(struct ds, row_log_str, DS_EMPTY_INITIALIZER);
+
 struct ovsdb_txn *
 ovsdb_txn_create(struct ovsdb *db)
 {
@@ -472,6 +479,41 @@ update_ref_counts(struct ovsdb_txn *txn)
     return for_each_txn_row(txn, check_ref_count);
 }
 
+static void
+ovsdb_txn_row_log(const struct ovsdb_txn_row *txn_row)
+{
+    static struct vlog_rate_limit rl_row_log = VLOG_RATE_LIMIT_INIT(30, 60);
+
+    if (!txn_row->table->log) {
+        return;
+    }
+
+    size_t n_columns = shash_count(&txn_row->table->schema->columns);
+    struct ovsdb_row *log_row;
+    const char *op = NULL;
+
+    if (!txn_row->old && txn_row->new) {
+        log_row = txn_row->new;
+        op = "inserted";
+    } else if (txn_row->old && txn_row->new
+               && !bitmap_is_all_zeros(txn_row->changed, n_columns)) {
+        log_row = txn_row->new;
+        op = "updated";
+    } else if (txn_row->old && !txn_row->new) {
+        log_row = txn_row->old;
+        op = "deleted";
+    }
+
+    if (op && !VLOG_DROP_INFO(&rl_row_log)) {
+        struct ds *ds = row_log_str_get();
+        ds_clear(ds);
+        ds_put_format(ds, "table:%s,op:%s,", txn_row->table->schema->name,
+                      op);
+        ovsdb_row_to_string(log_row, ds);
+        VLOG_INFO("%s", ds_cstr(ds));
+    }
+}
+
 static struct ovsdb_error *
 ovsdb_txn_row_commit(struct ovsdb_txn *txn OVS_UNUSED,
                      struct ovsdb_txn_row *txn_row)
@@ -495,6 +537,7 @@ ovsdb_txn_row_commit(struct ovsdb_txn *txn OVS_UNUSED,
         }
     }
 
+    ovsdb_txn_row_log(txn_row);
     ovsdb_txn_row_prefree(txn_row);
     if (txn_row->new) {
         txn_row->new->n_refs = txn_row->n_refs;
@@ -571,6 +614,8 @@ add_weak_ref(const struct ovsdb_row *src, const struct ovsdb_row *dst_,
     weak->column_idx = column->index;
     hmap_node_nullify(&weak->dst_node);
     ovs_list_push_back(ref_list, &weak->src_node);
+
+    n_weak_refs++;
 }
 
 static void
@@ -688,12 +733,13 @@ assess_weak_refs(struct ovsdb_txn *txn, struct ovsdb_txn_row *txn_row)
                 ovs_list_init(&weak->src_node);
             }
         }
-        ovsdb_datum_sort_unique(&deleted_refs, column->type.key.type,
-                                               column->type.value.type);
+        ovsdb_datum_sort_unique(&deleted_refs, &column->type);
 
         /* Removing elements that references deleted rows. */
-        ovsdb_datum_subtract(datum, &column->type,
-                             &deleted_refs, &column->type);
+        if (deleted_refs.n) {
+            ovsdb_datum_subtract(datum, &column->type,
+                                 &deleted_refs, &column->type);
+        }
         ovsdb_datum_destroy(&deleted_refs, &column->type);
 
         /* Generating the difference between old and new data. */
@@ -706,7 +752,7 @@ assess_weak_refs(struct ovsdb_txn *txn, struct ovsdb_txn_row *txn_row)
                                           &txn_row->old->fields[column->index],
                                           datum, &column->type);
             } else {
-                ovsdb_datum_clone(&added, datum, &column->type);
+                ovsdb_datum_clone(&added, datum);
             }
         }
 
@@ -716,8 +762,7 @@ assess_weak_refs(struct ovsdb_txn *txn, struct ovsdb_txn_row *txn_row)
                                &txn_row->added_refs, &deleted_refs, &zero);
         if (deleted_refs.n) {
             /* Removing all the references that doesn't point to valid rows. */
-            ovsdb_datum_sort_unique(&deleted_refs, column->type.key.type,
-                                                   column->type.value.type);
+            ovsdb_datum_sort_unique(&deleted_refs, &column->type);
             ovsdb_datum_subtract(datum, &column->type,
                                  &deleted_refs, &column->type);
             ovsdb_datum_destroy(&deleted_refs, &column->type);
@@ -957,9 +1002,17 @@ static struct ovsdb_error * OVS_WARN_UNUSED_RESULT
 count_atoms(struct ovsdb_txn *txn, struct ovsdb_txn_row *txn_row)
 {
     struct ovsdb_table *table = txn_row->table;
+    size_t n_columns = shash_count(&table->schema->columns);
     ssize_t n_atoms_old = 0, n_atoms_new = 0;
     struct shash_node *node;
 
+    if (txn_row->old && txn_row->new
+        && bitmap_is_all_zeros(txn_row->changed, n_columns)) {
+        /* This row didn't change, it has nothing to contribute
+         * to atom counters. */
+        return NULL;
+    }
+
     SHASH_FOR_EACH (node, &table->schema->columns) {
         const struct ovsdb_column *column = node->data;
         const struct ovsdb_type *type = &column->type;
@@ -1014,6 +1067,7 @@ ovsdb_txn_precommit(struct ovsdb_txn *txn)
      * was really a no-op. */
     error = for_each_txn_row(txn, determine_changes);
     if (error) {
+        ovsdb_txn_abort(txn);
         return OVSDB_WRAP_BUG("can't happen", error);
     }
     if (ovs_list_is_empty(&txn->txn_tables)) {
@@ -1067,7 +1121,7 @@ ovsdb_txn_precommit(struct ovsdb_txn *txn)
 }
 
 static struct ovsdb_txn*
-ovsdb_txn_clone(const struct ovsdb_txn *txn)
+ovsdb_txn_clone_for_history(const struct ovsdb_txn *txn)
 {
     struct ovsdb_txn *txn_cloned = xzalloc(sizeof *txn_cloned);
     ovs_list_init(&txn_cloned->txn_tables);
@@ -1084,14 +1138,22 @@ ovsdb_txn_clone(const struct ovsdb_txn *txn)
         struct ovsdb_txn_row *r;
         HMAP_FOR_EACH (r, hmap_node, &t->txn_rows) {
             size_t n_columns = shash_count(&t->table->schema->columns);
+
+            if (r->old && r->new
+                && bitmap_is_all_zeros(r->changed, n_columns)) {
+                /* The row is neither old or new and no columns changed.
+                 * No need to store in the history. */
+                continue;
+            }
+
             struct ovsdb_txn_row *r_cloned =
                 xzalloc(offsetof(struct ovsdb_txn_row, changed)
                         + bitmap_n_bytes(n_columns));
 
             r_cloned->uuid = r->uuid;
             r_cloned->table = r->table;
-            r_cloned->old = r->old ? ovsdb_row_clone(r->old) : NULL;
-            r_cloned->new = r->new ? ovsdb_row_clone(r->new) : NULL;
+            r_cloned->old = r->old ? ovsdb_row_datum_clone(r->old) : NULL;
+            r_cloned->new = r->new ? ovsdb_row_datum_clone(r->new) : NULL;
             memcpy(&r_cloned->changed, &r->changed, bitmap_n_bytes(n_columns));
             hmap_insert(&t_cloned->txn_rows, &r_cloned->hmap_node,
                         uuid_hash(&r_cloned->uuid));
@@ -1129,7 +1191,7 @@ ovsdb_txn_add_to_history(struct ovsdb_txn *txn)
 {
     if (txn->db->need_txn_history) {
         struct ovsdb_txn_history_node *node = xzalloc(sizeof *node);
-        node->txn = ovsdb_txn_clone(txn);
+        node->txn = ovsdb_txn_clone_for_history(txn);
         ovs_list_push_back(&txn->db->txn_history, &node->node);
         txn->db->n_txn_history++;
         txn->db->n_txn_history_atoms += txn->n_atoms;
@@ -1153,9 +1215,10 @@ ovsdb_txn_complete(struct ovsdb_txn *txn)
 
 /* Applies 'txn' to the internal representation of the database.  This is for
  * transactions that don't need to be written to storage; probably, they came
- * from storage.  These transactions shouldn't ordinarily fail because storage
- * should contain only consistent transactions.  (One exception is for database
- * conversion in ovsdb_convert().) */
+ * from storage or from relay.  These transactions shouldn't ordinarily fail
+ * because storage should contain only consistent transactions.  (One exception
+ * is for database conversion in ovsdb_convert().)  Transactions from relay
+ * should also be consistent, since relay source should have verified them. */
 struct ovsdb_error * OVS_WARN_UNUSED_RESULT
 ovsdb_txn_replay_commit(struct ovsdb_txn *txn)
 {
